## Here we list a bunch of rules and dependencies that can be used
## to build programs that link against the libtree and libsw libraries.
## This file used to be called Make.generic.  Now it's called Make.app,
## We keep a link from Make.generic to prevent unnecessary breakage.

ifndef special_rule_for_all
all: All
.PHONY:  All
endif

## But first, some things that are common to both library and application
## building.
include $(treedir)/Make-common/Make.shared

ifeq ($(PAROS),$(defaultPAROS))
# we don't need mpmy if we are using the default PAROS!
LIBDEPENDS=$(libdir)/libsw.a
else
LIBDEPENDS=$(libdir)/mpmy_$(PAROS)$(objsuf) $(libdir)/libsw.a
endif

# If we're going to the trouble to list out the full directory names
# in LIBDEPENDS, there's not much point in making the linker figure them
# out (possibly erroneously) again with -L... -l...
LOADLIBES_:=$(PRELIBS) $(LIBDEPENDS) -lm $(LOADLIBES)
# For the benefit of makefiles that wish to set LOADLIBES_ explicitly:
LIBMPMY = $(libdir)/mpmy_$(PAROS)$(objsuf)
LIBTREE = $(libdir)/libooctree.a
LIBSW = $(libdir)/libsw.a

obj-with-dir:=$(patsubst %.o,$(objdir)/%$(objsuf),$(obj))

### LINK.o is a GNUmake builtin, but it uses $(CC) instead of $(LINKCC)
### We are explicitly modifying the default rule so that we have more
### control over the actual link command.  $(LINKCC) should have been
### set in either Make.$(ARCH) or Make.default
LINK.o = $(LINKCC) $(LDFLAGS) $(TARGET_ARCH)

prog-with-arch:=$(programname).$(ARCH)$(ARCHsuf)$(PAROSsuf)
.PHONY : $(programname)
$(programname): $(prog-with-arch)
$(prog-with-arch) : $(objdir-exists) $(obj-with-dir) $(LIBDEPENDS)
	$(LINK.o) -o $@ $(obj-with-dir) $(LOADLIBES_)
	$(finishlink)

ifndef special_rule_for_all
All: $(prog-with-arch)
endif

# We sometimes like to write our asm using #defines:
%.s : %.S
	$(CPP) $< > $@

# A rules to guarantee that a $(objdir) exists.
$(objdir-exists) :
	if [ ! -d Objfiles ]; then mkdir Objfiles; fi
	if [ ! -d $(objdir) ]; then mkdir $(objdir); fi
	touch $(objdir-exists)

## This sed expr gets applied to the Makefile itself as part of the
## Make depends rule.  It converts foo.o into $(objdir)/foo.o

depends_sed_expr:='/DELETE THIS LINE/,$$s@\(.*\):@$$(objdir)/\1 :@'
