#include <string.h>
#include "Assert.h"
#include "Msgs.h"
#include "Malloc.h"
#include "mpmy.h"

#include "page.h"

#include "buckets.h"

#define GROWBY 10

void
BucketCreate(bucket_t *bk){
  bk->npgs = 0;
  bk->npgs_allocated = 0;
  bk->pages = NULL;
  bk->bytes_per_page = NULL;
  bk->ptr = NULL;
  bk->space = 0;
}

void
BucketDestroy(bucket_t *bk){
  int b;
  for(b=0; b<bk->npgs; b++){
    PgFree(bk->pages[b]);
  }
  Free(bk->pages);
  Free(bk->bytes_per_page);
  bk->npgs = 0;
  bk->pages = NULL;
  bk->bytes_per_page = NULL;
  bk->npgs_allocated = 0;
  bk->ptr = NULL;
}

void BucketToMem(bucket_t *bk, void *location){
  int b;

  for(b=0; b<bk->npgs; b++){
    PgHndl h = PgRef(bk->pages[b]);
    void *from = PgAddr(h);
    Msgf(("BtoM:  copying %d bytes from %p to %p\n", bk->bytes_per_page[b],
	  from, location));
    memcpy(location, from, bk->bytes_per_page[b]);
    PgUnref(h, PG_CLEAN);
    location = (char *)location + bk->bytes_per_page[b];
  }
}

void BucketAddData(bucket_t *bk, void *from, size_t sz){
  if( sz > bk->space ){
    int toppg;

    if( sz > PgSize )
      Error("BucketAddData(sz (%ld) > PgSize (%ld)\n", (long)sz, (long)PgSize);
    if( bk->ptr )
      PgUnref(bk->hndl, PG_DIRTY);
    toppg = bk->npgs++;
    if( toppg >= bk->npgs_allocated ){
      int newalloc = bk->npgs_allocated + GROWBY;
      bk->pages = Realloc(bk->pages, newalloc*sizeof(PgId));
      bk->bytes_per_page = Realloc(bk->bytes_per_page, newalloc*sizeof(int));
      bk->npgs_allocated = newalloc;
    }
    bk->pages[toppg] = PgAlloc();
    bk->bytes_per_page[toppg] = 0;
    bk->hndl = PgRef(bk->pages[toppg]);
    bk->ptr = PgAddr(bk->hndl);
    bk->space = PgSize;
  }

  memcpy(bk->ptr, from, sz);
  bk->space -= sz;
  bk->ptr = (char*)bk->ptr + sz;
  bk->bytes_per_page[bk->npgs-1] += sz;
}

void BucketLastItemOnPage(bucket_t *bk){
  if( bk->ptr ){
    PgUnref(bk->hndl, PG_DIRTY);
  }
  bk->ptr = NULL;
  bk->space = 0;
}

size_t BucketSize(bucket_t *bk){
  int b;
  size_t ret = 0;

  for(b=0; b<bk->npgs; b++){
    ret += bk->bytes_per_page[b];
  }
  return ret;
}



