/* Traverse a tree and call TreeChkSibGroup for every cell.  This verifies:
    - all bodies belong in the cell
    - bodies are stored in increasing key order.
    - bodies do not overlap with daughters.
    - daughters are in increasing key order.
    - no more than m bodies with the same 'child' number.

  The key of a body is not stored by the tree itself, so the traversal
  callback must "know" about the body.

  This version only traverses local (i.e., pgid.proc==MPMY_Procnum()) tcells.

  It would also be nice to somehow verify that all processors see
  identical trees in parallel environment.  We could skip the test on
  pgid.proc and apply some kind of checksum incrementally as the data
  are scanned.  Then a final reduction would tell us whether everybody
  had the same thing.  */

#include <stdlib.h>
#include "Msgs.h"
#include "mpmy.h"

#include "physics.h"

#include "page.h"
#include "tree.h"
#include "physics.h"
#include "traverse.h"

static void 
chkVisitSg(const sib_group *sg){
  Key_t *bkeys; 
  tbody *tbods;
  int i;

  Msgf(("Checking sg(key=%s)\n", PrintKey(sg->key)));
  bkeys = alloca(sg->ntbody*sizeof(Key_t));
  tbods = Tbodies(sg);
  for(i=0; i<sg->ntbody; i++){
    bkeys[i] = KeyFromFloats(tbods[i].pos, &UniverseBbox, nbodybits, key_order);
  }
  TreeChkSibGroup(sg, bkeys, NDIM, occupancy);
}

static int
chkMac(Key_t key, const tcell *tc){
  return ( tc->daughter.pgid.proc == MPMY_Procnum() ) ?
    MAC_REJECT : MAC_ACCEPT;

}

void
chktree(oocptr root){
  TraverseSg(root, chkMac, chkVisitSg);
  PagePollTillDone();
}
