#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>

#include "protos.h"
#include "Msgs.h"
#include "error.h"

#include "fiddledirs.h"

#include "mpmy.h"

void FiddleDirs(const char *newpath_fmt, int procnum){
  FILE *msgfp;
  char name[128];
  char newdir[PATH_MAX+1];

#if 0
  {
    char cwd[1024];
    getcwd(cwd, 1024);
    printf("Proc=%d, cwd=%s\n", MPMY_Procnum(), cwd);
  }
#endif

  sprintf(newdir, newpath_fmt, procnum);
  if( strlen(newdir) ){
    struct stat sb;
    errno=0;
    /* First stat the newdir to see if it exists and is a directory */
    if( stat(newdir, &sb) < 0 ){
      if( errno != ENOENT )
	Error("Can't stat(\"%s\").  errno: %d\n", newdir, errno);
      errno = 0;
      if( mkdir(newdir, 0777) ){
	if( errno == EEXIST ){
	  /* Stat said the file didn't exist (ENOENT), but mkdir fails
	     with EEXIST.  What's happening?   A symlink to a non-existent
	     location */
	  if( lstat(newdir, &sb) < 0){
	    Error("Total mystery.  You can't get here!\n");
	  }
	  if( !S_ISLNK(sb.st_mode) ){
	    Error("Not a symlink, mkdir says EEXIST, stat fails!\n");
	  }
	  /* We could try to do readlink (a BSD4.4ism, supported in Linux)
	     here, then do the mkdir at the end of the link.  That would
	     work for the 'standard' model of how we expect to use
	     this file.   But it's just too much trouble... */
	  Error("You have a symlink at %s, but it doesn't point anywhere!\n",
		newdir);
	}else{
	  Error("Could not mkdir(\"%s\").  errno: %d\n", newdir, errno);
	}
      }
      errno = 0;
      /* Now stat it again just to make sure */
      if( stat(newdir, &sb) < 0 ){
	Error("Could not stat(\"%s\") after mkdir!.  errno=%d\n", newdir, errno);
      }
    }
    if( !S_ISDIR(sb.st_mode) ){
      Error("Not a directory:  %s, mode=%#0o\n", newdir, sb.st_mode);
    }

    errno = 0;
    if( chdir(newdir) ){
      Error("Can't chdir(\"%s\"). errno: %d\n", newdir, errno);
    }
  }

  sprintf(name, "msgs.%d", procnum);
  msgfp = fopen(name, "w");
  Msg_addfile(msgfp, (Msgvfprintf_t)vfprintf, (Msgfflush_t)fflush);
  Msg_flushalways(1);

  /* A quick hack.  Leave stdout alone on node 0, which does the
   right thing in a uniprocessor env., and is close enough to right
   under lsv... */
  if( procnum > 0 ){
    sprintf(name, "stdout.%d", procnum);
    freopen(name, "w", stdout);
  }
  setlinebuf(stdout);
  printf("Hello world on processor %d\n", procnum);
}

