#include <string.h>
#include <math.h>
#include "singlio.h"
#include "physics.h"
#include "integrate.h"
#include "mpmy.h"
#include "protos.h"

#include "vop.h"

/* Called when a new acceleration has been computed.  Assume that
   the position and acceleration are at the 'current' time and the
   velocity is 'retarded' by half a timestep.  There's lots of
   opportunity for additional features/complexity here */
void
IntegrateBody(body *bp){
  /* First compute some gravdiags with the velocity synchronized with
     the positions */
  /* Push the velocities and positions forward */
  VV(bp->vel, += dt*bp->acc);
  VV(bp->pos, += dt*bp->vel);
}

/* Called after the body has been integrated to its new position,
   this function tries to update a UniverseBbox so that the body
   won't leave on the NEXT timestep. */
void
newUniverse(body *bp){
}

void
DiagsZero(gravdiags_t *diag){
  if( diag == NULL )
    return;

  memset(diag, 0, sizeof(*diag));
}

void
DiagsAdd(gravdiags_t *diag, body *bp, float veldtadjust){
  double vel[NDIM];
  if(diag == NULL )
    return;

  VVV(vel, = bp->vel, + veldtadjust*bp->acc);
  diag->ke += 0.5 * bp->mass * Dot(vel, vel);
  diag->pe += 0.5 * bp->mass * bp->phi;
  VV(diag->cmpos, += bp->mass * bp->pos);
  VV(diag->cmvel, += bp->mass * vel);
  VV(diag->cmacc, += bp->mass * bp->acc);
  diag->mtot += bp->mass;
  diag->nterms += bp->nterms;
}

void
DiagsFinish(gravdiags_t *diag){
  gravdiags_t tmp;

  if( diag == NULL )
    return;
  tmp = *diag;
  MPMY_Combine(&tmp, diag, sizeof(tmp)/sizeof(double), MPMY_DOUBLE, MPMY_SUM);
  if( diag->mtot > 0. ){
    VS(diag->cmpos, /= diag->mtot);
    VS(diag->cmvel, /= diag->mtot);
    VS(diag->cmacc, /= diag->mtot);
  }else{
    SinglShout("Mtot zero in DiagsFinish - not dividing through by mtot\n");
  }
}

void
DiagsPrint(gravdiags_t *diag){
  singlPrintf("Total Mass: %g, KE: %g, PE: %g, K+U: %g, nterms: %g\n",
	 diag->mtot, diag->ke, diag->pe, diag->ke+diag->pe, diag->nterms);
  singlPrintf("cmpos: " Sinfix("%g", " ") " norm: %g\n", 
	      Vinfix(diag->cmpos, COMMA), sqrt(Dot(diag->cmpos, diag->cmpos)));
  singlPrintf("cmvel: " Sinfix("%g", " ") " norm: %g\n",
	      Vinfix(diag->cmvel, COMMA), sqrt(Dot(diag->cmvel, diag->cmvel)));
  singlPrintf("cmacc: " Sinfix("%g", " ") " norm: %g\n",
	      Vinfix(diag->cmacc, COMMA), sqrt(Dot(diag->cmacc, diag->cmacc)));
}
