#include <stdlib.h>

#include "physics.h"

#include "key.h"
#include "tree.h"
#include "traverse.h"
#include "decomp.h"
#include "Msgs.h"

Key_t kSrch(double target, oocptr oocsg, int ndim,
	    double (*cCost)(Key_t, tcell *), 
	    void (*bVisit)(sib_group *, SplitTbodyInfo *)){
  double sum = 0.F;
  sib_group *sg;
  PgHndl sghndl;
  tcell *tc, *lasttc;
  unsigned int bits;
  unsigned int mask;
  int lev, maxlev, rshift;
  SplitTbodyInfo *tbinfo;
  int bnum;
  Key_t lastkey, sgkey;

  mask = (1<<ndim) - 1;
  sg = OOCFollow(oocsg, &sghndl);
  Msgf(("kSrch(%g, %s)\n", target, PrintSg(sg)));

  if( target == 0.F ){
    SeriousWarning("target=0.F in decompSrch, sg=%s\n", PrintSg(sg));
  }

  tbinfo = alloca(sg->ntbody * sizeof(*tbinfo));
  tc = &sg->tcells[0];
  lasttc = sg->tcells + sg->ncell;
  lev = TreeLevel(sg->key, ndim);
  bnum = 0;
  maxlev = ((KEYBITS-1)/ndim);
  if( lev >= 0 ){
    sgkey = sg->key;
  }else{
    /* it's the root cell with key=0x0 */
    Msgf(("lev<0, faking an sgkey=0x1\n"));
    sgkey = KeyInt(0x1);
    lev = 0;
  }
  rshift = ((maxlev - lev)-1)*ndim;
  Msgf(("lev=%d, rshift=%d, sgkey=%s\n", lev, rshift, PrintKey(sgkey)));

  (*bVisit)(sg, tbinfo);
  if( Msg_test(__FILE__) ){
    int i;
    for(i=0; i<sg->ntbody; i++){
      Msg_do("tbinfo[%d]={key=%s, value=%g}\n", 
	     i, PrintKey(tbinfo[i].key), tbinfo[i].value);
    }
  }
    
  lastkey = KeyLshift(sgkey, rshift + ndim);
  Msgf(("lastkey=%s\n", PrintKey(lastkey)));
  for(bits=0; bits<(1<<ndim); bits++){
    Key_t ret;
    float sumplus;
    if( tc<lasttc && tc->bits == bits ){
      /* The bits corresponds to the next tcell */
      sumplus = sum + (*cCost)(sgkey, tc);
      if( sumplus >= target ){
	ret = kSrch(target-sum, tc->daughter, ndim, cCost, bVisit);
	PgUnref(sghndl, PG_CLEAN);
	return ret;
      }
      sum = sumplus;
      tc++;
      lastkey = KeyOrInt( KeyLshift(sgkey, ndim), bits);
      lastkey = KeyLshiftFill( lastkey, rshift);
    }else{
      /* The bits (may) correspond to the next (few) body(s) */
      while( bnum<sg->ntbody && KeyAndInt(KeyRshift(tbinfo[bnum].key, rshift), mask)==bits ){
	sumplus = sum + tbinfo[bnum].value;
	Msgf(("bits=%d, sump (%g) = %g + %g\n", bits, sumplus, sum, tbinfo[bnum].value));
	if( sumplus >= target ){
	  Msgf(("We have a winner!  lastkey=%s, ", PrintKey(lastkey)));
	  Msgf(("tbinfo[%d].key=%s\n", bnum, PrintKey(tbinfo[bnum].key)));
	  Msgf(("interpolate to %g: ", (target-sum)/(sumplus-sum)));
	  ret = KeyInterpolate(lastkey, tbinfo[bnum].key, 
			       (target-sum)/(sumplus-sum));
	  Msgf(("result is %s\n", PrintKey(ret)));
	  PgUnref(sghndl, PG_CLEAN);
	  return ret;
	}
	sum = sumplus;
	lastkey = tbinfo[bnum].key;
	bnum++;
      }
    }
  }

  PgUnref(sghndl, PG_CLEAN);
  /* Is this an error? It can happen from roundoff */
  return KeyInt(0x0);
}
