#ifndef _PhySicsDotH_
#define _PhySicsDotH_

#define DEBUGGABLE_TBODY

#include "key.h"
#include "keycvt.h"
#include "timers.h"
#include "buckets.h"

/* Change the names of the user-types.  If you play this game, the this
   file MUST come before any #inclusion of "tree.h"  (including
   recursive inclusions).  This can be an annoyance, in which case the
   options are to a) live with the warnings about mismatched types or
   b) add ugly explicit casts to your callbacks. */
#define _tbody struct tbody_s
#define _cellDataIntermediate struct cofm_s
#define _cellDataFinal struct cell_s
struct tbody_s;
struct cofm_s;
struct cell_s;

#include "tree.h"
#include "decomp.h"
#include "page.h"

#define NDIM 3
/* typedefs for the physics structures */

typedef struct{
  float mass;
  float pos[NDIM];
  float vel[NDIM];
  float acc[NDIM];
  float phi;
  int ident;
  int nterms;
  Key_t key;
} body;

typedef struct cell_s{
  float mass;
  float pos[NDIM];
  float rcrit;
  float B2, B3, bmax;
  int daughters;
  float cost;
} cell;

/* This goes in the tree.  A "tree"body doesn't need any dynamical
   information like velocity, accel, pe, etc. */
typedef struct tbody_s{
  float mass;
  float pos[NDIM];
  float cost;
#ifdef DEBUGGABLE_TBODY
  /* Not necessary for physics.  Helpful for debugging! */
  int ident;
  Key_t key;
#endif
} tbody;

typedef struct cofm_s{
  float mass;
  float pos[NDIM];
  float B2, bmax;
  int daughters;
  float cost;
}cofm;  

/* In main.c */
extern tbbox UniverseBbox;
extern double Mtot;
extern oocptr srcroot;
extern int nbodybits;

extern bucket_t *bodies;	/* bods for the current timestep */
extern bucket_t *nextbodies;	/* where we put bods for the next timestep */
extern krange_t *bucketdividers;
extern krange_t *decompranges;

void doBodies( void (*callback)(const body *) );

/* In cmdline.c */
extern int gnobj;
extern float dt;
extern float errtol;		/* absolute, in force units */
extern float epsilon;
extern int key_order;		/* MORTON or PH */
extern int verbose;
extern int debug;
extern int bktspace;
extern int nbuckets;
extern float rcritfac;
extern int max_sinksplit_per_body;
extern int occupancy;

/* In print.c */
char *PrintCellContents(const cell *p);
char *PrintBodyContents(const body *p);
char *PrintTBodyContents(const tbody *p);
char *PrintCofmContents(const cofm *cmp);
void BHPrintTree(int localonly, int (*printf_like)(const char *, ...));
Key_t keyFromTbody(const _tbody *);

/* In cofm17.c */
extern BuildFuncs_t GravBfuncs;
Key_t keyFromTbody(const _tbody *_tb);
double cellCost(Key_t, tcell *cp);
void bodyCostVisit(sib_group *sg, SplitTbodyInfo *tbinfo);
double cellSz(Key_t, tcell *tc);
void bodySzs(sib_group *sg, SplitTbodyInfo *tbinfo);


#include "integrate.h"

/* In bh.c */
void BHTraverse(float eps, float GNewt, gravdiags_t *diags);
extern Counter_t BHMacCnt;
/* In bhvec.c */
void BHvTraverse(float eps, float GNewt, gravdiags_t *diags);

/* In dl.c */
extern Counter_t DLMacCnt, DLUncertainMacCnt, DLUncertainSplitSinkCnt;
extern Counter_t DLPeelCnt, DLDepthCnt;
extern Timer_t DLPollWaste;
void DLTraverse(float eps, float GNewt, gravdiags_t *diags);

/* In chktree.c */
extern void chktree(oocptr root);

/* In sort.c */
extern Timer_t SortTm;
void SortAndAddToTree(tree_t *tp);

#endif
