#ifndef _PlaceBodsdotH_
#define _PlaceBodsdotH_

/* Functions for 'placing' bodies for the next timestep. */

#include "key.h"
#include "buckets.h"

/* A high-level interface to 'buckets' that allows you to "fire and
  forget" about the placement of bodies.  If you are going to create a
  bunch of new objects all the same size, and you want them stored in
  an array of buckets, then this is the interface for you.  The bodies
  NEED NOT BE SORTED.  This is the interface you use when you're assembling
  buckets to be sorted later.

  PlaceBodsBegin Inputs:  
    - decompranges:  the key ranges that will be used to separate processors
  from one another (e.g., use decomp() in decomp.h).  
    - nbuckets: how many buckets
    - bucketdividers:  the  ranges to use to separate the buckets 
  within this processor (e.g.,  use splitKeys() in decomp.h).
    - unitsz: size (in bytes) of each object.

  PlaceBodsBegin will save some of this info in static storage.  At least
    for now, the system is not re-entrant, i.e., you can have only one
    Begin/End pair open at a time.

  PlaceBody:  give a pointer to the object and its associated key.  Fire and
    forget.

  PlaceBodsEnd:  Assert that no more bodies are on the way.  Buffers
    will be flushed, communication will happen, etc.  Upon completeion
    your buckets will be ready.

    */

void PlaceBodsBegin(int ndim, size_t unitsz, krange_t *decompranges, 
		    int nbuckets, bucket_t *tobuckets, krange_t *bucketdividers);
void PlaceBody(void *bp, Key_t key);
void PlaceBodsEnd(void);

#endif /* _PlaceBodsdotH_ */
