#include <stdio.h>		/* for sprintf */
#include "Assert.h"
#include "mpmy.h"
#include "physics.h"
#include "traverse.h"
#include "gc.h"
#include "verify.h"
#include "vop.h"
#include "protos.h"
#include "Msgs.h"

/* Make these return a ptr to static data so we can finesse the */
/* problem of what kind of FILE *! */
char *
PrintCellContents(const cell *p)
{
    static char contents_string[256];

    if( p == NULL){
	sprintf(contents_string, "\t(null)");
    }else{
	sprintf(contents_string, "\tmass: %.4g, nd:%d, rcrit: %.2g, cost: %.2g\n"
		"\t" Sinfix("%.4f", " "), 
		p->mass, p->daughters, p->rcrit, p->cost,
		Vinfix(p->pos, COMMA));
    }
    return contents_string;
}

char *
PrintBodyContents(const body *p)
{
    static char contents_string[256];

    if( p == NULL){
	sprintf(contents_string, "\t(null)");
    }else{
	sprintf(contents_string, "\tid:%d, mass:%.4g, key=%s\n"
		"\t" Sinfix("%.4f", " "),
		p->ident, p->mass, PrintKey(p->key),
		Vinfix(p->pos, COMMA));
    }
    return contents_string;
}

/* For out of bits confirmation */
char *
PrintBodyContentsLong(const body *p)
{
    static char contents_string[256];

    if( p == NULL){
	sprintf(contents_string, "\t(null)");
    }else{
	sprintf(contents_string, "id: %d @%p mass: %7.4g, key: %s\n"
		"\tpos: " Sinfix("%g", " ")
		" vel: " Sinfix("%g", " ")
		"\n\tacc: " Sinfix("%g", " ")
		" phi: %g, nterms: %d\n",
		p->ident, p, p->mass, PrintKey(p->key),
		Vinfix(p->pos, COMMA),
		Vinfix(p->vel, COMMA),
		Vinfix(p->acc, COMMA),
		p->phi, p->nterms);
    }
    return contents_string;
}

char *
PrintTBodyContents(const tbody *p)
{
    static char contents_string[256];
    Key_t key;

    key = KeyFromFloats(p->pos, &UniverseBbox, nbodybits, key_order);
    if( p == NULL){
	sprintf(contents_string, "\t(null)");
    }else{
	sprintf(contents_string, "\tkey: %s mass:%.4g, cost:%.4g\n"
		"\t" Sinfix("%.4f", " "),
		PrintKey(key), p->mass, p->cost,
		Vinfix(p->pos, COMMA));
    }
    return contents_string;
}

char *PrintCofmContents(const cofm *cmp){
    static char ret[512];
    sprintf(ret, "Br: mass: %.3g, ndaughters=%d, pos=(%.3f %.3f %.3f), bmax:%.2g\n",
	   cmp->mass, cmp->daughters, 
	    cmp->pos[0], cmp->pos[1], cmp->pos[2],
	    cmp->bmax);
    return ret;
}

static int printLocalOnly;
static int (*printf_like)(const char *fmt, ...);

static void 
GravPrintBodies(const sib_group *sg){
  _tbody *_tb = Tbodies(sg);
  tbody *tbods = (tbody *)_tb;
  int i;

  (*printf_like)("Bodies (%d) in sg %s\n", sg->ntbody, PrintKey(sg->key));
  for(i=0; i<sg->ntbody; i++){
    (*printf_like)("body %d: %s\n", i, PrintTBodyContents(&tbods[i]));
  }
}

static int 
GravPrintMAC(Key_t key, const tcell *tc){
  cell *cp = Contents(tc);

  (*printf_like)("cell@%p key=%s.%o %s\n", 
	 cp, PrintKey(key), tc->bits, PrintCellContents(cp));
  if ( printLocalOnly && tc->daughter.pgid.proc != MPMY_Procnum() ){
    (*printf_like)("Remote daughter at %s+%#x\n", 
		   PrintPgId(tc->daughter.pgid), tc->daughter.offset);
    return MAC_ACCEPT;
  }else{
    return MAC_REJECT;
  }
}

void 
BHPrintTree(int localonly, int (*printf_like_arg)(const char *, ...) ){
  printf_like = printf_like_arg;
  printLocalOnly = localonly;
  TraverseSg(srcroot, GravPrintMAC, GravPrintBodies);
  PagePollTillDone();
}
