#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "physics.h"
#include "tree.h"
#include "cmdline.h"

#include "timers.h"
#include "vop.h"
#include "Malloc.h"
#include "Msgs.h"
#include "mpmy.h"

Timer_t SortTm;

static int
cmpBodyKeys(const void *_b1, const void *_b2){
  const body *b1 = (const body *)_b1;
  const body *b2 = (const body *)_b2;

  return KeyCmp(b1->key, b2->key);
}

void
SortAndAddToTree(tree_t *tp){
  int b;
    /* loop over buckets */
    for(b=0; b<nbuckets; b++){
      bucket_t *bkp;
      size_t bsz;
      int nbod;
      body *base;
      void *inmemp;
      int p, i;

      bkp = &bodies[b];
      bsz = BucketSize(bkp);
      nbod = bsz/sizeof(body);
      base = Malloc(bsz);
      StartTimer(&SortTm);
      Msgf(("Calling BucketToMem(base=%p)\n", base));
      BucketToMem(bkp, (void *)base);
      Msgf(("Calling qsort(nbod=%d)\n", nbod));
      if( nbod > 0 )
	qsort(base, nbod, sizeof(body), cmpBodyKeys);
      /* Write the data back out.  This may not be optimal.  It might
	 be better to just sort it again when we want it sorted... */
      inmemp = base;
      for(p=0; p<bkp->npgs; p++){
	PgId pg = bkp->pages[p];
	PgHndl hndl = PgRef(pg);
	void *bktaddr = PgAddr(hndl);
	int nbytes = bkp->bytes_per_page[p];
	
	memcpy(bktaddr, inmemp, nbytes);
	inmemp = (char *)inmemp + nbytes;
	PgUnref(hndl, PG_DIRTY);
      }
      StopTimer(&SortTm);

      /* And within the bucket (now in memory), loop over bodies */
      for(i=0; i<nbod; i++){
	body *bp; 
	tbody tb;
	bp = &base[i];
	tb.mass = bp->mass;
	tb.cost = (float)bp->nterms/(float)gnobj;
#ifdef DEBUGGABLE_TBODY
	tb.ident = bp->ident;
	tb.key = bp->key;
#endif
	VV(tb.pos, = bp->pos);
	Msg("createbodies", ("Body(%d:%d:%d) %s\n",
	      MPMY_Procnum(), b, i, PrintBodyContents(bp)));
	TreeBuildAdd(tp, bp->key, &tb, sizeof(tbody));
      }
      printf("Processor %d, bkt %d, nbod=%d\n", MPMY_Procnum(), b, nbod);

      Free(base);
    }
}
