#ifndef _do_Grav_dotH
#define _do_Grav_dotH

#if (NDIM != 3)
# error do_grav only works with NDIM 3
#endif

/* All of the 'p' and 'end' should really be Ivec_t *, but I can't 
   be bothered to change them all... */
typedef struct {
    union{
	double not_used;
	struct {
	    float mass;
	    float pos[NDIM];
	} used;
    }u;
} Ivec_t;

/* Call this routine unless you are REALLY sure you know what you're doing */
void
do_grav(const Ivec_t *p, const Ivec_t *end, const float *pos0, float *mass0, 
	float *acc0, float *phi0, const float *eps2p, int *ncut);

/* Generic, C-coded routine.  #undef UNROLL */
void 
do_grav_u1(const float *p, const float *end, const float *pos0, float *mass0, 
	float *acc0, float *phi0, const float *eps2p, int *ncut);

/* C version, #define UNROLL 2 */
void 
do_grav_u2(const float *p, const float *end, const float *pos0, float *mass0, 
	float *acc0, float *phi0, const float *eps2p, int *ncut);

/* C version, uses the chebychev approx, #define UNROLL 2 */
void 
do_grav_cheb(const float *p, const float *end, const float *pos0, float *mass0, 
	float *acc0, float *phi0, const float *eps2p, int *ncut);

void 
do_grav_cheb_unroll2(const float *p, const float *end, const float *pos0, float *mass0, 
	float *acc0, float *phi0, const float *eps2p, int *ncut);

void 
do_grav_cheb_unroll3(const float *p, const float *end, const float *pos0, float *mass0, 
	float *acc0, float *phi0, const float *eps2p, int *ncut);

/* Fast asm for i860, #define UNROLL 3 */
void 
do_grav_fast(const float *p, const float *end, const float *pos0, float *mass0, 
	float *acc0, float *phi0, const float *eps2p, int *ncut);

/* Fast asm for cm5, #define UNROLL 32 */
void 
do_grav_cm5(const float *p, const float *end, const float *pos0, float *mass0, 
	float *acc0, float *phi0, const float *eps2p, int *ncut);

#endif
