#ifndef __PageDoTh_
#define __PageDoTh_

#include "timers.h"
#include "page-internal.h"

#ifdef __cplusplus
extern "C"{
#endif /* __cplusplus */

/* Must be 0 and 1 or the code in PgUnref breaks! */
enum PgHygiene_e{PG_CLEAN=0, PG_DIRTY=1};

typedef inmeminfo *PgHndl;

/* Initialize the whole shebang */
void PgInit(int lgpgsz, int wsetpages, const char *filename);
/* Terminate the whole shebang */
void PgTerm(void);
/* Reference a page -- i.e., guarantee that it's in memory until it's
   unref'ed.  The PgHndl return value can be used in PgUnref, or in
   PgAddr to get the address. */
PgHndl PgRef(PgId pgid);
/* Increment the reference count on a page -  Call PgReref if you already
   have a PgHndl, and for some reason we want to guarantee that it
   stays around, but cannot avoid calling PgUnref.  (e.g., you're inside
   a callback in a tree traversal). */
void PgReref(PgHndl hndl);
/* Unreference a page.  These accumulate.  Tell us if it got dirtied while
   it was locked, which determines whether it needs to be written out */
void PgUnref(PgHndl hndl, enum PgHygiene_e dirtied);

/* Flush all the remote pages.  Too sweeping, but it works for now */
void PgFlushRemote(void);

/* Print a pgid */
char *PrintPgId(PgId pgid);

/* Ask about the status of a page.  Return logical OR of these: */
#define PG_STAT_REMOTE 0x1
#define PG_STAT_INMEM 0x2
#define PG_STAT_REFED 0x4
#define PG_STAT_DIRTY 0x8
int PgStatus(PgId pgid);

/* How big are the pages (an arg to PgInit) */
extern int PgSize;
/* values of some accumulated stats. Should these be Counter_t's too? */
extern int PgNrefcnt;			/* how many are referenced right now */
extern int PgReuse;			/* how many locks were in-mem */

extern Timer_t PgInTm, PgOutTm, RemoteTm;
extern Timer_t PgPollTm, PgPollTillDoneTm;
extern Timer_t PgPollLoopTm;
extern Counter_t PgInCnt, PgOutCnt;
extern Counter_t PgRemoteInCnt, PgRemoteSendCnt;
extern Counter_t PgPollLoopCnt;

/* These are in pagemanage.c.  THey actually know nothing at all about
   the internals of the paging library, but why bother with another .h file? */
PgId PgAlloc(void);
void PgFree(PgId);
void PgStats();			/* send some stats to singlPrintf */
extern int PgInUse;		/* how many pages are currently allocated */
extern int PgHwm;		/* high-water-mark of allocated pages. */

extern void PagePollTillDone(void);
extern void PagePoll(void);

typedef struct{
  PgId pgid;
  unsigned int offset;
} oocptr;

#define OOCFOLLOW(ooc, hndl, ptr) \
  do{\
       oocptr _ooc = (ooc);\
       hndl=PgRef((_ooc).pgid);\
       ptr=(void *)( ((char *)PgAddr(hndl)) + (_ooc).offset );\
  }while(0)

#if defined(__GNUC__) || defined(page5DotC)

#undef INLINE
#if defined (__GNUC__)
 #if !defined (page5DotC)
  #define INLINE extern __inline__
 #else
  #define INLINE __inline__  /* inline, but make an external symbol too */
 #endif
#else /* not __GNUC__, but page5DotC */
 #define INLINE
#endif

INLINE void *PgAddr(PgHndl hndl){
  return hndl->base;
}

INLINE void *OOCFollow(oocptr ooc, PgHndl *hndl){
  *hndl=PgRef(ooc.pgid);
  return (void *)( ((char *)PgAddr(*hndl)) + ooc.offset );
}

/* Don't we need PgRef and PgUnref inlined too!? */

#endif /* __GNUC__ */

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif
