/* Create a hash table that allows us to look up 'PgHndl's
   with a PgId key.  
 
   At this point, don't worry about fancy abstractions that allow us
   to look up generic data with generic keys.  This is still probably
   too complicated...

   Use the Dll data objects because we have them and they aren't too
   mismatched to the task at hand.  It could probably be lighter-weight,
   but this is a good place to start.
*/

#ifndef _MyHasHdotH
#define _MyHasHdotH

#include "page.h"
#include "dll.h"
#include "chn.h"

/* The entire hash table. */
typedef struct{
  Dll *tbl;
  unsigned int tblsz;
  Chn chn;
} hash_t;

/* An entry in the hash table. */
typedef struct {
  inmeminfo *data;
  PgId id;
  Dll_elmt *myelmt;
  Dll *mydll;
} hash_entry_t;

/* State variables that allow us to iterate over the contents of the
   hash table */
typedef struct {
  hash_t *ht;
  Dll *lastdll;
  Dll_elmt *last;
} hash_iterator_t;

/* Create a table.  */
void hash_make(hash_t *, int nentries);

/* Destroy a hash table */
void hash_destroy(hash_t *);

/* Perform internal consistency checks on a hash table */
int hash_check(hash_t *);

/* Look for the key, and if not found, create a new entry */
hash_entry_t *hash_get(hash_t *, PgId *);

/* Look for the key, and if not found, return NULL */
hash_entry_t *hash_find(hash_t *, PgId *);

/* Delete an entry */
void hash_delete(hash_t *, hash_entry_t*);

/* release any 'locks' on the entry.  If we aren't running multi-threaded
   then this does nothing.  There's no harm in keeping the API, though */
void hash_release(hash_t *, hash_entry_t*);

/* Start an iteration through the hash table.  By calling hash_iter_next()
   repeatedly, all entries will be returned.

   If multithreaded, hold a lock on the entire table based on the
   value of hold_global_lock.  Use it if you want to be absolutely
   sure nobody is modifying the table while you're looking at it.  If
   a global lock is not held, then entries may be added or deleted by
   other threads while the iteration proceeds.  Consistency is a
   complicated issue.  Do we care?  */
void hash_iter_start(hash_iterator_t *iterator, hash_t *ht, int hold_global_lock);
void hash_iter_finish(hash_iterator_t *iterator);

/* Return the next hash entry in the iteration.  Return NULL when
   there are no more entries.

   If multi-threaded, A per-entry lock is held until the next call to
   hash_iter_next().  */
hash_entry_t *hash_iter_next(hash_iterator_t *);



#endif /* myHasHdotH */
