#ifndef _Tree_InternaldotH_
#define _Tree_InternaldotH_

#include "chn.h"

/* Tree data structures, macros and functions that we don't want the
   casual user to see or use.  Typically, this file will be included by
   tree.h, so there is no real 'protection' enforcable. At
   least the casual user will be able to tell at a glance which functions
   are for him.  */

#ifdef __cplusplus
extern "C"{
#endif /* __cplusplus */

/* DUMMY_ARRAY_LENGTH: the array length for dummy arrays.  Gcc allows 0-length
 arrays for precisely this purpose.  ANSI forbids them... */
#if defined(__GNUC__) &&0/* gdb on linux seems unhappy with 0-length arrays */

#define DUMMY_ARRAY_LENGTH 0
#else
#define DUMMY_ARRAY_LENGTH 1
#endif

/* TC_TMP: This is a temporary object that is used during tree
   construction.  Size isn't too important.  There are somthing like
   O(logN) of them in memory at any one time.  The 'contents' is
   allocated and deallocated by the library, and the user Aggregate_t
   routine is called with a pointer to it, along with information about
   how much space is there.
   */
typedef struct{
  char bits;			/* various flags, etc. including 
				   low NDIM bits of the key*/
  _cellDataIntermediate *contents;
  size_t contents_sz;
  struct sg_tmp_s *daughter;
  struct sg_tmp_s *sibs;	/* could be an 'offset' instead */
  oocptr oocdaughter;		/* points to a sib_group */
} tc_tmp;

/* SG_TMP:  This is the under-construction sib_group.  Again, there are
   never really that many of them in memory, so size isn't crucial. */
typedef struct sg_tmp_s{
  tc_tmp *parent;
  Key_t key;			/* NOT left-shifted.  low bits not 0. */
  int flags;			/* COMPLETE_HI, COMPLETE_LO */
  unsigned char ncell;			/* how many cells?  */
  unsigned char ntbody;			/* how many tbodies? */
  tc_tmp tc_tmps[1<<MAXNDIM];		/* really tc_tmps[ncell] */
  /* There's a lot of waste here! */
  _tbody *tbodyptrs[MAXM<<MAXNDIM]; /* really tbodies[ntbody] */
  size_t tbodysz[MAXM<<MAXNDIM];   
} sg_tmp;

/* Used to hold bodies for sewing data sets in different processors
   back together.  Basically, it just records the args to TreeBuildAdd */
struct body_info{
  _tbody *tb;
  size_t sz;
  Key_t key;
} ;

/* BUILDSTATE_S: internal state of tree build process.  It has meaning
   only from TreeBuildStart through TreeBuildEnd */
struct buildstate_s{
};

/* TREE_T: The tree structure itself. */
typedef struct tree_s{
  oocptr rootsg;		/* points to the root sib_group */
  int ndim;			/* Dimensionality of space. */
  int m;			/* maximum terminal occupancy. */
  sg_tmp *curr_grp;
  sg_tmp rootsgtmp;
  sg_tmp *firstsg;
  int curr_child;
  int curr_level, maxlevs;
  tc_tmp *next_tc_tmp;
  int rshift;
  int ndimmask;
  int count;
  Key_t Key0;
  Key_t prevkey;
  struct BuildFuncs_s *Bfuncs;
  struct body_info *holding, *orphans;
  int norphans;
  int holdidx;
  int holdingsz;
  Chn sgchn;
  Chn CofmChn;
  int npages;
  int maxpages;
  PgId *pages;			/* which pages are mine? */
} tree_t;

/* These will be fully elaborated in tree.h */
struct sib_group_s;
struct tcell_s;
struct BuildFuncs_s;

#define COMPLETE_HI (0x1)
#define COMPLETE_LO (0x2)
#define Completep(sg) ((sg->flags & (COMPLETE_HI|COMPLETE_LO)) == (COMPLETE_HI|COMPLETE_LO))

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* _Tree_InternaldotH_ */
