/*
 * Copyright 1991-1996 Michael S. Warren and John K. Salmon.  All Rights Reserved.
 */

#ifdef __STDC__
#define GLUE(a,b) a##b
#else
#define GLUE(a,b) a/**/b
#endif

/* UGLY, but you can use these for parentheses inside the VV macros! */
#define LPAREN (
#define RPAREN )
#define COMMA ,

/* The infix functions can't be handled with a switch */
#if NDIM == 3
#define Sinfix(s, op) s op s op s
#define Vinfix(a, op) a[0] op a[1] op a[2]
#define VVinfix(a, b, op) a[0] b[0] op a[1] b[1] op a[2] b[2]
#define Vxinfix(a, op) GLUE(a,0) op GLUE(a,1) op GLUE(a,2)
#define VxVxinfix(a, b, op) GLUE(a,0) GLUE(b,0) op GLUE(a,1) GLUE(b,1) op GLUE(a,2) GLUE(b,2)
#elif NDIM == 2
#define Sinfix(s, op) s op s
#define Vinfix(a, op) a[0] op a[1]
#define VVinfix(a, b, op) a[0] b[0] op a[1] b[1]
#define Vxinfix(a, op) GLUE(a,0) op GLUE(a,1)
#define VxVxinfix(a, b, op) GLUE(a,0) GLUE(b,0) op GLUE(a,1) GLUE(b,1)
#elif NDIM == 1
#define Sinfix(s, op) s
#define Vinfix(a, op) a[0]
#define VVinfix(a, b, op) a[0] b[0]
#define Vxinfix(a, op) GLUE(a,0)
#define VxVxinfix(a, b, op) GLUE(a,0) GLUE(b,0)
#endif

/* The infix operators can be used in, e.g.,:
   printf(Sinfix("%g", " "), Vinfix(pos, COMMA)) 
   The macros below may inspire other ideas... */

#define Dot(a, b) (VVinfix(a, *b, +))

#define Vdecl(type, v) type v[NDIM]

#define Dotx(a, b) (VxVxinfix(a, *b, +))

/* Use for declarations.  We always declare up to subscript 3 because
 the switch switch statements have (dead) code that refers to all three
 elements.
*/
/* #define Vxd(a) Vxinfix(a,;) BROKEN */
#define Vxd(a) GLUE(a,0); GLUE(a,1); GLUE(a,2)

/* Use in prototypes */
/* #define Vxp(a) Vxinfix(a,COMMA) BROKEN */
#define Vxp(a) GLUE(a,0), GLUE(a,1), GLUE(a,2)

#define VS(a, b) do{ switch(NDIM){ \
  case 3: a[2] b; \
  case 2: a[1] b; \
  case 1: a[0] b; \
      }}while(0)

#define VVS(a, b, s) do { switch(NDIM){ \
        case 3: a[2] b[2] s; \
        case 2: a[1] b[1] s; \
        case 1: a[0] b[0] s; \
      }}while(0) 

#define VV(a, b) do{ switch(NDIM) { \
  case 3: a[2] b[2]; \
  case 2: a[1] b[1]; \
  case 1: a[0] b[0]; \
      }}while(0)

#define VVV(a, b, c) do{ switch(NDIM) { \
    case 3: a[2] b[2] c[2]; \
    case 2: a[1] b[1] c[1]; \
    case 1: a[0] b[0] c[0]; \
      }}while(0)

#define VVVV(a, b, c, d) do{ switch(NDIM){\
        case 3: a[2] b[2] c[2] d[2]; \
        case 2: a[1] b[1] c[1] d[1]; \
        case 1: a[0] b[0] c[0] d[0]; \
      }}while(0) 


#define VVVS(a, b, c, s) do{ switch(NDIM){\
        case 3: a[2] b[2] c[2] s; \
        case 2: a[1] b[1] c[1] s; \
        case 1: a[0] b[0] c[0] s; \
      }}while(0) 


#define VVVVV(a, b, c, d, e) do{ switch(NDIM){\
        case 3: a[2] b[2] c[2] d[2] e[2]; \
        case 2: a[1] b[1] c[1] d[1] e[1]; \
        case 1: a[0] b[0] c[0] d[0] e[0]; \
      }}while(0) 

#define VVVVVS(a, b, c, d, e, s) do{ switch(NDIM){\
        case 3: a[2] b[2] c[2] d[2] e[2] s; \
        case 2: a[1] b[1] c[1] d[1] e[1] s; \
        case 1: a[0] b[0] c[0] d[0] e[0] s; \
      }}while(0) 

#define VVVVVV(a, b, c, d, e, f) do{ switch(NDIM){\
        case 3: a[2] b[2] c[2] d[2] e[2] f[2]; \
        case 2: a[1] b[1] c[1] d[1] e[1] f[1]; \
        case 1: a[0] b[0] c[0] d[0] e[0] f[0]; \
      }}while(0) 

/* Eleven args! */
#define Velv(a, b, c, d, e, f, g, h, i, j, k) do{ switch(NDIM){\
        case 3: a[2] b[2] c[2] d[2] e[2] f[2] g[2] h[2] i[2] j[2] k[2]; \
        case 2: a[1] b[1] c[1] d[1] e[1] f[1] g[1] h[1] i[1] j[1] k[1]; \
        case 1: a[0] b[0] c[0] d[0] e[0] f[0] g[0] h[0] i[0] j[0] k[0]; \
      }}while(0) 

#define VxS(a, b) do{ switch(NDIM){\
	case 3: GLUE(a,2) b; \
	case 2: GLUE(a,1) b; \
	case 1: GLUE(a,0) b; \
      }}while(0) 

#define VxV(a, b) do{ switch(NDIM){\
        case 3: GLUE(a,2) b[2]; \
        case 2: GLUE(a,1) b[1]; \
        case 1: GLUE(a,0) b[0]; \
      }}while(0) 

#define VVx(a, b) do{ switch(NDIM){\
        case 3: a[2] GLUE(b,2); \
        case 2: a[1] GLUE(b,1); \
        case 1: a[0] GLUE(b,0); \
      }}while(0) 

#define VxVx(a, b) do{ switch(NDIM){\
        case 3: GLUE(a,2) GLUE(b,2); \
        case 2: GLUE(a,1) GLUE(b,1); \
        case 1: GLUE(a,0) GLUE(b,0); \
      }}while(0) 


#define VxVV(a, b, c) do{ switch(NDIM){\
        case 3: GLUE(a,2) b[2] c[2]; \
        case 2: GLUE(a,1) b[1] c[1]; \
        case 1: GLUE(a,0) b[0] c[0]; \
      }}while(0) 

#define VxVVS(a, b, c, s) do{ switch(NDIM){\
        case 3: GLUE(a,2) b[2] c[2] s; \
        case 2: GLUE(a,1) b[1] c[1] s; \
        case 1: GLUE(a,0) b[0] c[0] s; \
      }}while(0) 

#define VxVVVS(a, b, c, d, s) do{ switch(NDIM){\
        case 3: GLUE(a,2) b[2] c[2] d[2] s; \
        case 2: GLUE(a,1) b[1] c[1] d[1] s; \
        case 1: GLUE(a,0) b[0] c[0] d[0] s; \
      }}while(0) 

#define VxVVx(a, b, c) do{ switch(NDIM){\
        case 3: GLUE(a,2) b[2] GLUE(c,2); \
        case 2: GLUE(a,1) b[1] GLUE(c,1); \
        case 1: GLUE(a,0) b[0] GLUE(c,0); \
      }}while(0) 

#define VxVxV(a, b, c) do{ switch(NDIM){\
        case 3: GLUE(a,2) GLUE(b,2) c[2]; \
        case 2: GLUE(a,1) GLUE(b,1) c[1]; \
        case 1: GLUE(a,0) GLUE(b,0) c[0]; \
      }}while(0) 

#define VxVxVx(a, b, c) do{ switch(NDIM){\
        case 3: GLUE(a,2) GLUE(b,2) GLUE(c,2); \
        case 2: GLUE(a,1) GLUE(b,1) GLUE(c,1); \
        case 1: GLUE(a,0) GLUE(b,0) GLUE(c,0); \
      }}while(0) 

#define VVxVx(a, b, c) do{ switch(NDIM) { \
	case 3: a[2] GLUE(b,2) GLUE(c,2); \
	case 2: a[1] GLUE(b,1) GLUE(c,1); \
	case 1: a[0] GLUE(b,0) GLUE(c,0); \
      }}while(0) 
