#define KEYdotC
/* Most of the definitions are in key.h */
#include <stdio.h>		/* just for sprintf */
#include "key.h"
#include "protos.h"
#include "Msgs.h"

/* Non-inlined definitions go here. */

char *
PrintKey(Key_t key)
{
    static char str[128];

#ifndef LONG_LONG_KEYS
    if (NK == 1)
      sprintf(str, "%lo", key.k[0]);
    else
      /* This only works for NDIM==3 */
      sprintf(str, "%010lo.%01lo%010lo", key.k[1] >> 1, 
	      ((key.k[1] & 01) << 2) | (key.k[0] >> 30),
	      key.k[0] & ~(3U << 30));
#else
    sprintf(str, "%010o.%01o%010o", 
	    (unsigned long int)(key.k[0]>>33), /* bits 63-33 */
	    (unsigned long int)((key.k[0]>>30) & 7), /* bits 32,31,30 */
	    (unsigned long int)(key.k[0] & ((1<<30)-1))); /* 29-0 */
#endif
		
    return str;
}

int
TreeLevel(Key_t key, int ndim)
{
    int level;
    int chubits = (KEYBITS-1)/ndim;
    Key_t testkey;

    /* First check whether it's a 'body' (at the deepest level.)  */
    /* This will save considerable time... */
    testkey = KeyLshift(KeyInt(1), chubits*ndim);
    if( KeyEQ( testkey, KeyAnd(testkey, key) ) )
	return chubits;

    /* Now start looking from low levels */
    testkey = KeyInt(1);
    for (level = 0; level<chubits; level++){
	if( KeyEQ(key, testkey) )
	    return level;
	key = KeyRshift(key, ndim);
    }
    
    return -1;
}

int
isBodyKey(Key_t key, int ndim)
{
    int chubits = (KEYBITS-1)/ndim;
    Key_t testkey;

    testkey = KeyLshift( KeyInt(0x1), chubits*ndim);
    return KeyEQ( testkey, KeyAnd(testkey, key));
}

/* Find the common level between two 'body'-keys */
int
CommonLev(Key_t bkey1, Key_t bkey2, int ndim)
{
    int level = (KEYBITS-1)/ndim;
    Key_t key0 = KeyInt(0);

    for (bkey1 = KeyXOR(bkey1, bkey2); KeyNEQ(bkey1, key0); level--)
      bkey1 = KeyRshift(bkey1, ndim);
    
    return(level);
}

int
KeyContained(Key_t outer, Key_t key, int ndim)
{
    int ret;
    int difference;
    
    /* What if difference is negative?! */
    difference = TreeLevel(key, ndim) - TreeLevel(outer, ndim);
    key = KeyRshift(key, ndim*difference);
    ret = KeyEQ(key, outer);
    return(ret);
}

/* Should we have a treekeys.h for
   key functions that rely on the specific structure of cell/body keys? 
   Return true if the boxes signified by the two keys 'overlap' */
int
KeyOverlap(Key_t key1, Key_t key2, int ndim){
  /* Asks whether key1 and key2 overlap.  Both are cell keys. */
  int lev1,  lev2;
  lev1 = TreeLevel(key1, ndim);
  lev2 = TreeLevel(key2, ndim);

  if( lev1 > lev2 ){
    key1 = KeyRshift(key1, ndim*(lev1 - lev2));
  }else if(lev1 < lev2){
    key2 = KeyRshift(key2, ndim*(lev2 - lev1));
  }
  return KeyEQ(key1, key2);
}

/* Interpolate between two keys */
#if NK == 2
Key_t KeyInterpolate(Key_t lok, Key_t hik, double frac){
  _KTYPE idiff;
  Key_t kdiff;
  unsigned int shift;

  kdiff = KeySub(hik, lok);
  shift = 0;
  /* Binary search could do this in 5 iterations instead of 32... */
  while( kdiff.k[1] ){		/* kdiff[1] are the high bits */
    shift++;
    kdiff = KeyRshift(kdiff, 1);
  }
  idiff = kdiff.k[0];
  
  /* Now do arithmetic with the shifted difference, to get the right
     fraction.  Is rounding right here? */
  idiff *= frac;
  /* Finally, shift back left, and add it back to lok */
  return KeyAdd(lok, KeyLshift(KeyInt(idiff), shift));
}

#elif NK==1

Key_t KeyInterpolate(Key_t lok, Key_t hik, double frac){
  Key_t kdiff;

  kdiff.k[0] = lok.k[0] + frac*(hik.k[0]-lok.k[0]);
  return kdiff;
}

#endif
