#define PHKEYFROMINTS _PHKeyFromInts2d
#define INTSFROMPHKEY _IntsFromPHKey2d
#define NDIM 2

/* The possible places to start. */
/* They can only have an even number of bits turned on. */
#define S00 0
#define S11 1

/* one for each of the possible startindices */
static int sindex_to_mask[1<<(NDIM-1)] = {0, 3};
static int smask_to_index[1<<NDIM] = {S00, -1, -1, S11};

/* The possible "path-types" (there are NDIM of them) */
#define Px 0
#define Py 1

static int pindex_to_mask[NDIM] = {2, 1};

/* What are the large transitions in each of the NDIM basic paths? */
static int bigtrans[NDIM][1<<NDIM] = {
    {Py, Px, Py, Px}, /* Px */
    {Px, Py, Px, Py} /* Py */
};

/* What are the small transitions on each of the three basic paths? */
static int ltltrans[NDIM][1<<NDIM] = {
    {Py, Px, Px, Py}, /* Px */
    {Px, Py, Py, Px} /* Py */
};


#include "peanond.c"

#ifdef STANDALONE
/* Compile this source file with -DSTANDALONE and link with the
   usual tree library to get a quick-and-dirty program that generates
   a postscript PH curve.
*/

#define MAXDEPTH (15/NDIM)
#include <stdlib.h>
#include <stdio.h>

#define SZ 6. /* inches */
int
main(int argc, char **argv){
    int depth;
    unsigned int i, ikey;
    unsigned int ik[NDIM];
    int type, start;
    unsigned int revk[NDIM], revd;
    Key_t key, base;

    depth = atoi(argv[1]);
    start = atoi(argv[2]);
    type = atoi(argv[3]);
    if( depth < 0 )
	Error("bad depth\n");
    if( type < 0 || type >= NDIM )
	Error("bad type\n");
    if( start < 0 || start >= 1<<(NDIM-1) )
	Error("bad start");

    Warning("This is test of the emergency warning system\n");
    /* Test each ikey in a 3-d grid.  Make sure that 
       _IntsFromKey(KeyFromInts(ik)) == ik  */
    for(ik[0]=0; ik[0]<(1<<depth); ik[0]++){
	for(ik[1]=0; ik[1]<(1<<depth); ik[1]++){
	  key = PHKEYFROMINTS(ik, depth, start, type);
	  revd = INTSFROMPHKEY(key, revk, depth, start, type);
	  assert(revd == depth );
	  if( revk[0]!=ik[0] || revk[1]!=ik[1])
	    Warning("ik=(%x %x) -> %s -> revk=(%x %x)\n",
			   ik[0], ik[1],
			   PrintKey(key),
			   revk[0], revk[1]);
	}
    }

    /* Test each key starting at 0. Make sure that 
       KeyFromInts(IntsFromKey(k))==k */
    printf("%%!\n");
    printf("/L {2 copy lineto stroke moveto pop} def\n");
    printf("72 72 translate\n");
    printf("%%scale to 6 inches width, hgt\n");
    printf("%g %g scale\n", 72.*SZ/(1<<depth), 72.*SZ/(1<<depth));
    printf("%% line width of 1pt\n");
    printf("%g setlinewidth\n", (1<<depth)/(SZ*72));
    printf("0 0 moveto\n");

    base = KeyLshift(KeyInt(1), depth*NDIM);
    i=0;
    for(ikey=0; ikey < (1<<(NDIM*depth)); ikey++){
      Key_t revkey;
      unsigned int di[NDIM];
      unsigned int iklast[NDIM];
      int d2;

      revd = INTSFROMPHKEY(base, ik, depth, start, type);
      assert(revd == depth);
      revkey = PHKEYFROMINTS(ik, depth, start, type);
      if( KeyNEQ(revkey, base)){
	char s[64];
	strcpy(s, PrintKey(revkey));
	Warning("key %s -> %x %x -> %s\n",
		       PrintKey(base), ik[0], ik[1], s);
      }
      printf("%d %d %d L\n", i, ik[0], ik[1]);
      i++;
      VVV(di, = ik, - iklast);
      d2 = Dot(di, di);
      if( d2 != 1 && ikey ){
	Warning("Move by more than 1 at ikey=%#0o\n", ikey);
      }

      VV(iklast, = ik);
      base = KeyAddInt(base, 1);
    }
    printf("showpage\n");
    exit(0);
}
#endif /* STANDALONE */
