#define PHKEYFROMINTS _PHKeyFromInts3d
#define INTSFROMPHKEY _IntsFromPHKey3d
#define NDIM 3

/* The possible places to start. */
/* They can only have an even number of bits turned on. */
#define S000 0
#define S011 1
#define S101 2
#define S110 3
/* one for each of the possible startindices */
static int sindex_to_mask[1<<(NDIM-1)] = {0, 3, 5, 6};
static int smask_to_index[1<<NDIM] = {S000, -1, -1, S011, -1, S101, S110, -1};

/* The possible "path-types" (there are NDIM of them) */
#define Px 0
#define Py 1
#define Pz 2

static int pindex_to_mask[NDIM] = {4, 2, 1};

/* What are the large transitions in each of the three basic paths? */
static int bigtrans[NDIM][1<<NDIM] = {
    {Py, Pz, Py, Px, Py, Pz, Py, Px}, /* Px */
    {Px, Pz, Px, Py, Px, Pz, Px, Py}, /* Py */
    {Px, Py, Px, Pz, Px, Py, Px, Pz} /* Pz */
};

/* What are the small transitions on each of the NDIM basic paths? */
static int ltltrans[NDIM][1<<NDIM] = {
    {Py, Pz, Pz, Px, Px, Pz, Pz, Py}, /* Px */
    {Px, Pz, Pz, Py, Py, Pz, Pz, Px}, /* Py */
    {Px, Py, Py, Pz, Pz, Py, Py, Px} /* Pz */
};

#include "peanond.c"

#ifdef STANDALONE
#include <stdio.h>
#include <stdlib.h>

#define MAXDEPTH (15/NDIM)
/* Compile this source file with -DSTANDALONE and link with the
   usual tree library to get a quick-and-dirty program that generates
   runs a consistency check on the 3d integer/phkey conversions.
*/

int
main(int argc, char **argv){
    int depth;
    unsigned int ikey;
    unsigned int ik[NDIM];
    int type, start;
    unsigned int revk[NDIM], revd;
    Key_t key, base;

    depth = atoi(argv[1]);
    start = atoi(argv[2]);
    type = atoi(argv[3]);
    if( depth < 0 )
	Error("bad depth\n");
    if( type < 0 || type >= NDIM )
	Error("bad type\n");
    if( start < 0 || start >= 1<<(NDIM-1) )
	Error("bad start");

    Warning("This is a test of the warning system.\nHad this been an actual warning,\nyou would have been advised of its nature\n");
    /* Test each ikey in a 3-d grid.  Make sure that 
       _IntsFromKey(KeyFromInts(ik)) == ik  */
    for(ik[0]=0; ik[0]<(1<<depth); ik[0]++){
	for(ik[1]=0; ik[1]<(1<<depth); ik[1]++){
	    for(ik[2]=0; ik[2]<(1<<depth); ik[2]++){
		key = PHKEYFROMINTS(ik, depth, start, type);
		revd = INTSFROMPHKEY(key, revk, depth, start, type);
		assert(revd == depth );
		if( revk[0]!=ik[0] || revk[1]!=ik[1] || revk[2]!=ik[2] )
		    Warning("ik=(%x %x %x) -> %s -> revk=(%x %x %x)\n",
				   ik[0], ik[1], ik[2],
				   PrintKey(key),
				   revk[0], revk[1], revk[2]);
	    }
	}
    }
    printf("passed intsFromKey(KeyFromInts(ik)) == ik test\n");

    /* Test each key starting at 0. Make sure that 
       KeyFromInts(IntsFromKey(k))==k */
    base = KeyLshift(KeyInt(1), depth*NDIM);
    for(ikey=0; ikey < (1<<(NDIM*depth)); ikey++){
      Key_t revkey;
      unsigned int di[NDIM];
      unsigned int iklast[NDIM];
      int d2;

      revd = INTSFROMPHKEY(base, ik, depth, start, type);
      assert(revd == depth);
      revkey = PHKEYFROMINTS(ik, depth, start, type);
      if( KeyNEQ(revkey, base)){
	char s[64];
	strcpy(s, PrintKey(revkey));
	Warning("key %s -> %x %x %x -> %s\n",
		       PrintKey(base), ik[0], ik[1], ik[2], s);
      }
      VVV(di, = ik, - iklast);
      d2 = Dot(di, di);
      if( d2 != 1 && ikey ){
	Warning("Move by more than 1 at ikey=%#0o\n", ikey);
      }
	
      VV(iklast, = ik);
      base = KeyAddInt(base, 1);
    }
    printf("passed KeyFromInts(IntsFromKey(k))==k (and continuity) test\n");
    exit(0);
}

#endif /* STANDALONE */
