/* This file is #included in another source file which defines
   NDIM and declares various arrays.  It's the easiest way of
   supporting multiple dimensions for the peano-hilbert stuff.
*/

#include "Msgs.h"
#include "Assert.h"
#include "key.h"
#include "vop.h"

static int bitmap[1<<(NDIM-1)][NDIM][1<<NDIM]; /* range 1<<NDIM */
static int revbitmap[1<<(NDIM-1)][NDIM][1<<NDIM]; /* range 1<<NDIM */
static int startmap[1<<(NDIM-1)][NDIM][1<<NDIM]; /* range 1<<(NDIM-1) */
static int typmap[1<<(NDIM-1)][NDIM][1<<NDIM]; /* range NDIM */
static int setup_done;

static void setup(void){
    unsigned int start, typ, j;
    unsigned int smap[1<<NDIM], bmap[1<<NDIM], lmap[1<<NDIM];
    unsigned int bt, lt, bm, st;

    for(typ=0; typ<NDIM; typ++){
	for(start=0; start<(1<<(NDIM-1)); start++){
	    bm = st = sindex_to_mask[start];
	    for(j=0; j<(1<<NDIM); j++){
		smap[j] = smask_to_index[st];
		lmap[j] = ltltrans[typ][j];
		bmap[j] = bm;
		bt = pindex_to_mask[ bigtrans[typ][j] ];
		lt = pindex_to_mask[ ltltrans[typ][j] ];
		bm = bm^bt;
		st = st^(bt^lt);
	    }
	    /* This little wierdness arranges that map works in the */
	    /* right direction...really */
	    for(j=0; j<(1<<NDIM); j++){
		bm = bmap[j];
		startmap[start][typ][bm] = smap[j];
		bitmap[start][typ][bm] = j;
		revbitmap[start][typ][j] = bm; /* a shot in the dark */
		typmap[start][typ][bm] = lmap[j];
	    }
	}
    }
    setup_done = 1;
}

Key_t PHKEYFROMINTS(unsigned int ikey[NDIM], unsigned int depth, int start, int type){
    Key_t ret;
    unsigned int bits;
    unsigned int rshift;
    unsigned int otype;
    Vxd(unsigned int ikey);
    ret = KeyInt(1);

    if( !setup_done )
	setup();

    Msgf(("phkeyfromints%dd(" Sinfix("%#x", ",") "), depth=%d, start=%d, type=%d\n",
	  NDIM, Vinfix(ikey, COMMA), depth, start, type));
    VxV(ikey, = ikey);
    for(rshift=depth; rshift; ){
	rshift--;
	bits = (ikey0>>rshift)&1;
#if NDIM > 1
	bits |= ((ikey1>>rshift)&1) << 1;
#if NDIM > 2
	bits |= ((ikey2>>rshift)&1) << 2;
#if NDIM > 3
	# error NDIM>3 code not written!
#endif
#endif
#endif
	ret = KeyOrInt(KeyLshift(ret, NDIM), bitmap[start][type][bits]);
	otype = type;
	type = typmap[start][otype][bits];
	start = startmap[start][otype][bits];
    }
    Msgf(("return %s\n", PrintKey(ret)));
    return ret;
}

/* Convert from a PH key to a NDIM-tuple of ints. */
/* Return the "depth" of the key */

unsigned int 
INTSFROMPHKEY(Key_t key, unsigned int ikey[NDIM], 
	  int depth, int start, int type){
    unsigned int lobits, unscrambled;
    unsigned int otype;
    unsigned int rshift, ret;
    Vxd(unsigned int out);
    Key_t keymax;
    Key_t key0;

    if( !setup_done )
	setup();
    keymax = KeyLshift(KeyInt(1), NDIM*depth);
    key0 = KeyInt(0);

    VxS(out, = 0);
    /* We need to start at the left, so we need to figure out where the */
    /* left of key is! */
    while( KeyEQ(KeyAnd(keymax, key), key0) ){
	keymax = KeyRshift(keymax, NDIM);
	depth--;
    }
    ret = depth;

    rshift = depth*NDIM;
    Msgf(("IntsFromPHKey%dd(%s, depth=%d, start=%d, type=%d)\n", 
	  NDIM, PrintKey(key), depth, start, type));
    while( rshift > 0 ){
        Key_t kr;
	rshift -= NDIM;
	depth--;
	kr = KeyRshift(key, rshift);
	lobits = KeyAndInt(kr, (1<<NDIM)-1);
	Msgf(("rshift=%d, kr=%s, lobits: %d\n", rshift, PrintKey(kr), lobits));
	unscrambled = revbitmap[start][type][lobits];
	
	out0 |= (unscrambled&1)<<depth;
#if NDIM > 1
	out1 |= ((unscrambled>>1)&1)<<depth;
#if NDIM > 2
	out2 |= ((unscrambled>>2)&1)<<depth;
#endif
#endif
	otype = type;
	type = typmap[start][otype][unscrambled];
	start = startmap[start][otype][unscrambled];
    }
    VVx(ikey, = out);
    Msgf(("ikey=(" Sinfix("%#x", ",") ") return %d\n", Vinfix(ikey, COMMA), ret));
    return ret;
}

