#include <SDF.h>
#include <error.h>
#include <Malloc.h>
#include <gc.h>
#include <mpmy.h>
#include <math.h>
#include "readbodies.h"


#define AT_A_TIME 10000

char *ReadBodiesSDFhdr = NULL;

void
ReadBodies(const char *filename, int nnames, int sizeof_body,
	   struct name_address_s *name_address, void (*dobody)(/* body * */)){
    void *buf;
    int at_a_time;
    int i, n, nn,gnobj;
    int nnmax, nmax;
    SDF *sdfp;
    int *nns;
    char **names;
    void **addrs;
    int *starts;
    int start;
    int *sizes;

    if( filename == NULL ){
	return;
    }

    at_a_time = AT_A_TIME;
    buf = Malloc(sizeof_body*at_a_time);
    sdfp = SDFopen(ReadBodiesSDFhdr, filename);
    if( sdfp == NULL ){
	Error("Could not SDFopen(NULL, %s): %s\n", 
	      filename?filename:"<null>", 
	      SDFerrstring);
    }
    /* This could be parallelized by fiddling with the usual
       modular arithmetic in gnobj, nproc, etc. */
    gnobj = SDFnrecs(name_address[0].name, sdfp);
    NobjInitial(gnobj, MPMY_Nproc(), MPMY_Procnum(), &n, &start);

    /* Avoid a hang by keeping track of how many reads the busiest
       processor will do.  We keep doing reads until nmax have been
       read by the guy that plans to read nmax.  We might read 0 on
       our last iteration, but that's no problem.  This should be part
       of NobjInitial... */
    nmax = gnobj/MPMY_Nproc();
    if( gnobj%MPMY_Nproc() )
	nmax++;
    
    names = Malloc(nnames*sizeof(char *));
    nns = Malloc(nnames*sizeof(int));
    addrs = Malloc(nnames*sizeof(void *));
    sizes = Malloc(nnames*sizeof(int));
    starts = Malloc(nnames*sizeof(int));
    for(i=0; i<nnames; i++){
	names[i] = name_address[i].name;
	addrs[i] = (char *)buf + name_address[i].offset;
	sizes[i] = sizeof_body;
	starts[i] = start;
    }
    while(nmax>0){
	int ret;
	nn = ( n < at_a_time ) ? n : at_a_time;
	nnmax = (nmax < at_a_time) ? nmax : at_a_time;
	for(i=0; i<nnames; i++){
	    nns[i] = nn;
	}
	ret = SDFseekrdvecsarr(sdfp, nnames, names, starts, nns, addrs, sizes);
	if( ret != 0 ){
	    SeriousWarning("SDFrdvecs failed: %s\n", SDFerrstring);
	}
	for(i=0; i<nn; i++){
	    (*dobody)((char *)buf + i*sizeof_body);
	}
	n -= nn;
	nmax -= nnmax;
	for(i=0; i<nnames; i++){
	    starts[i] += nn;
	}
    }
    SDFclose(sdfp);
    Free(sizes);
    Free(addrs);
    Free(nns);
    Free(names);
    Free(buf);
}

