/*
    SDF Library for reading Self-Describing Files
    Copyright (C) 1991,1992  John K. Salmon

    Terms and conditions are specified in the file "copyright.h",
    and more precisely in the file COPYING.LIB which you should have
    received with this library.
*/
/* 

This is a very basic "smoke test" for the SDF library.
This program ought to compile and run as: 
   prog SDFfilename
or 
   prog DataFilename SDFHdrFilename

If it does, that doesn't prove anything, but if it doesn't, 
then something is wrong.

*/

#include <stdio.h>
#include <stdlib.h>
#include "SDF.h"
#include "Msgs.h"
#include "protos.h"
#include "error.h"

#include "mpmy.h"

extern int SDFyydebug;

int main(int argc, char **argv){
    SDF *hdr;
    char *dataname, *hdrname;
    int i;
    char *name;
    char **vec_names;
    int nvecs;
    float x[32], y[32];
    struct { float vx, vy; } vels[32];
    char text[384];
    float GNewt;
    float Iter;
    int npart;
    enum SDF_type_enum type;

    MPMY_Init(&argc, &argv);
#ifdef DEBUG_MALLOC
    malloc_debug(2);
#endif
    if( strcmp(argv[1], "-d") == 0 ){
	argv++;
	SDFyydebug = 1;
	if( MPMY_Nproc() > 1 ){
	    char name[256];
	    sprintf(name, "misc.SDFtst/msg.%d", MPMY_Procnum());
	    MsgdirInit(name);
	}else{
	    Msg_addfile(stderr, vfprintf, fflush);
	}
	Msg_on("SDFfuncs.c");
	Msg_on("SDFhdrio.c");
	Msg_on("SDF");
	Msg_on("SDF-parse.y");
	Msg_on("SDF-lex.l");
	Msg_on("./SDFfuncs.c");
	Msg_on("./SDFhdrio.c");
	Msg_on("./SDF");
	Msg_on("./SDF-parse.y");
	Msg_on("./SDF-lex.l");
    }
    dataname = argv[1];
    if( argc > 2 )
	hdrname = argv[2];
    else
	hdrname = NULL;
    hdr = SDFopen(hdrname, dataname);
    if( argc > 3 ){
	int bufsz;
	bufsz = atoi(argv[3]);
	SDFsetmaxbufsz(bufsz);
    }

    if( hdr == NULL ){
	printf("Could not open files: %s\n", SDFerrstring);
	exit(1);
    }
    vec_names = SDFvecnames(hdr);
    nvecs = SDFnvecs(hdr);

    for(i=0; i<nvecs; i++){
	void *value;

	name = vec_names[i];
	type = SDFtype(name, hdr);
	value = malloc(SDFarrcnt(name, hdr)*SDFtype_sizes[type]);
	SDFseekrdvecs(hdr,
		      name, 0, 1, value, 0,
		      NULL);
	printf("Vector {%s %s[%d]}[%d]: ",
	       SDFtype_names[type],
	       name,
	       SDFarrcnt(name, hdr),
	       SDFnrecs(name, hdr));
	switch(type){
	case SDF_CHAR:
	    printf("%s...\n", (char *)value);
	    break;
	case SDF_INT:
	    printf("%d...\n", *(int *)value);
	    break;
	case SDF_FLOAT:
	    printf("%g...\n", *(float *)value);
	    break;
	case SDF_DOUBLE:
	    printf("%g...\n", *(double *)value);
	    break;
	default:
	    printf("Unrecognized type!\n");
	}
	free(value);
    }

    if( SDFhasname("tree_hdr_text", hdr) ){
	SDFrdvecs(hdr, 
		  "text", 1, &text[0], 1,
		  NULL);
	printf("Text is: %s\n", text);
    }

    if( SDFhasname("vx", hdr) && SDFhasname("vy", hdr) ){
	if( SDFseekrdvecs(hdr,
			  "vx", 16, 16, &vels[16].vx, sizeof(vels[0]),
			  "vy", 16, 16, &vels[16].vy, sizeof(vels[0]),
			  NULL) ){
	    printf("Couldn't read vx, vy: %s\n", SDFerrstring);
	    exit(2);
	}
	printf("Read the 16-31 vx,vy entries in the file with seekrd:\n");
	for(i=16; i<32; i++){
	    printf("%g %g\n", vels[i].vx, vels[i].vy);
	}
    }else{
	printf("This isn't a tree-like file, so further tests won't work.\n");
	exit(0);
    }

    if( SDFrdvecs(hdr,
	     "x", 32, &x[0], sizeof(x[0]),
	     "y", 32, &y[0], sizeof(y[0]),
	     NULL) ){
	printf("Couldn't read x, y: %s\n", SDFerrstring);
	exit(2);
    }

    SDFgetintOrDefault(hdr, "npart", &npart, 0);
    printf("Nbody = %d\n", npart);

    if( SDFseek("vx", 0, SDF_SEEK_SET, hdr) ){
	printf("Couldn't seek vx back to 0\n");
	exit(3);
    }
    if( SDFseek("vy", 0, SDF_SEEK_SET, hdr) ){
	printf("Couldn't seek vy back to 0\n");
	exit(4);
    }
    if( SDFrdvecs(hdr,
	     "vx", 16, &vels[0].vx, sizeof(vels[0]),
	     "iter", 1, &Iter, sizeof(Iter),
	     "Gnewt", 1, &GNewt, sizeof(GNewt),
	     "vy", 16, &vels[0].vy, sizeof(vels[0]),
	     NULL) ){
	printf("Couldn't read vx, npart, ...: %s\n", SDFerrstring);
	exit(2);
    }	

    printf("GNewt = %g\n", GNewt);
    printf("Iter = %g\n", Iter);

    printf("The first 32 entries in the file are:\n");
    printf("x y vx vy:\n");
    for(i=0; i<32; i++){
	printf("%g %g %g %g\n", x[i], y[i], vels[i].vx, vels[i].vy);
    }

    SDFclose(hdr);
#ifdef DEBUG_MALLOC
    malloc_print(stdout);
#endif
    exit(0);
    /* NOTREACHED */
    return 0;
}

