/* Like bench.c, but we are sending in both directions simultaneously. */
#include "mpmy.h"
#include "singlio.h"
#include "mpmy_time.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include "protos.h"

void *sendbuf, *recvbuf;

int main(int argc, char **argv){
  int lgmaxlen = 24;
  int lgminlen = 6;
  int lglen;
  int procnum, nproc;
  void *wctmr;
  double wctime;
  double meanwc;
  double maxwc;

  MPMY_Init(&argc, &argv);
  if( argc >= 2) 
    lgminlen = atoi(argv[1]);
  if( argc >= 3 )
    lgmaxlen = atoi(argv[2]);
  wctmr = MPMY_CreateTimer(MPMY_WC_TIME);
  procnum = MPMY_Procnum();
  nproc = MPMY_Nproc();

  sendbuf = malloc( 1<<lgmaxlen );
  recvbuf = malloc( 1<<lgmaxlen );

  if( procnum == (nproc-1) && nproc%2 ){
    printf("Proc: %d has no partner.  Exiting.\n", procnum);
    exit(0);
  }

  singlPrintf("#Shifting %d bytes in each test\n", 1<<lgmaxlen);
  singlPrintf("# Pktlength wctime wcmax bwidth\n");
  singlFflush();
  for( lglen = lgmaxlen; lglen >= lgminlen; lglen--){
    int len;
    int ntimes;

    MPMY_Sync();
    len = 1<<lglen;
    ntimes = 1<<(lgmaxlen - lglen);
    MPMY_ClearTimer(wctmr);
    MPMY_StartTimer(wctmr);
    while( ntimes-- ){
      MPMY_Shift(procnum^1, recvbuf, len, sendbuf, len, NULL);
    }
    MPMY_StopTimer(wctmr);
    MPMY_Combine(&wctime, &meanwc, 1, MPMY_DOUBLE, MPMY_SUM);
    MPMY_Combine(&wctime, &maxwc, 1, MPMY_DOUBLE, MPMY_MAX);

    meanwc /= nproc;
    singlPrintf("%12d %#.3g %#.3g %#.3g\n", 
	     len, meanwc, maxwc, (2<<lgmaxlen)/((1<<20)*meanwc));
    singlFflush();
  }
  singlPrintf("Bye!\n");
  exit(0);
}

