/* The simplest possible two-node program. */
#include "mpmy.h"
#include "singlio.h"
#include "mpmy_time.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include "protos.h"

#define MEGA (1000000.)

void *buf;

MPMY_Comm_request send_hndl, recv_hndl;

void
ping(int length, int partner, void *buf){
  /* The recv is already posted before entering */
  MPMY_Isend(buf, length, partner, 1001, &send_hndl);
  MPMY_Wait(send_hndl, NULL);
  MPMY_Wait(recv_hndl, NULL);
  MPMY_Irecv(buf, length, partner, 1001, &recv_hndl);
}

void
pong(int length, int partner, void *buf){
  MPMY_Wait(recv_hndl, NULL);
  MPMY_Isend(buf, length, partner, 1001, &send_hndl);
  MPMY_Wait(send_hndl, NULL);
  MPMY_Irecv(buf, length, partner, 1001, &recv_hndl);
}

#define INITIATOR (procnum < nproc/2)
#define PARTNER (procnum ^ (nproc/2))

int main(int argc, char **argv){
  int lgmaxlen = 24;
  int lgminlen = 6;
  int lgtotal;
  int lglen;
  int procnum, nproc, partner;
  void *wctmr;
  double wctime;
  double meanwc;
  double maxwc;

  MPMY_Init(&argc, &argv);
  if( argc >= 2) 
    lgminlen = atoi(argv[1]);
  if( argc >= 3 )
    lgmaxlen = atoi(argv[2]);
  wctmr = MPMY_CreateTimer(MPMY_WC_TIME);
  procnum = MPMY_Procnum();
  nproc = MPMY_Nproc();

  if( nproc%2 ){
    printf("Must be an even number of processors\n");
    exit(0);
  }

  buf = malloc( 1<<lgmaxlen );

  partner = PARTNER;

  singlPrintf("#Ping-ponging %d bytes in each test\n", 1<<lgmaxlen);
  singlPrintf("# Totlen Pktlength wctime wcmax bwidth(mean) bw(min) latency(mean) latency(max)\n");
  singlFflush();
  lgtotal = lgmaxlen;
  for( lglen = lgmaxlen; lglen >= lgminlen; lglen--){
    int len;
    int ntimes;

    len = 1<<lglen;
    ntimes = 1<<(lgtotal - lglen);
    MPMY_ClearTimer(wctmr);

    MPMY_Irecv(buf, len, partner, 1001, &recv_hndl);
    MPMY_Sync();
    MPMY_StartTimer(wctmr);
    
    while( --ntimes ){
      if( INITIATOR ){
	ping(len, partner, buf);
      }else{
	pong(len, partner, buf);
      }
    }
    MPMY_Isend(buf, len, partner, 1001, &send_hndl);
    MPMY_Wait(send_hndl, NULL);
    MPMY_Wait(recv_hndl, NULL);
    
    MPMY_StopTimer(wctmr);
    if( INITIATOR ){
      wctime = MPMY_ReadTimer(wctmr);
    }else{
      wctime = 0.0;
    }
    MPMY_Combine(&wctime, &meanwc, 1, MPMY_DOUBLE, MPMY_SUM);
    MPMY_Combine(&wctime, &maxwc, 1, MPMY_DOUBLE, MPMY_MAX);

    meanwc /= (nproc/2);
    ntimes = 1<<(lgtotal - lglen);
    /* bandwidth is estimated as 2*totalbytes/time, counting the bytes
       that move in each direction.  We also count it as two separate
       packets for the msecperpacket estimate */
    singlPrintf("%9d %9d %#7.3g %#7.3g %#7.3g %#7.3g %7.3g %7.3g\n", 
		1<<lgtotal, len, meanwc, maxwc, 
		(2<<lgtotal)/(MEGA*meanwc),
		(2<<lgtotal)/(MEGA*maxwc),
		meanwc*1000./(2*ntimes),
		maxwc*1000./(2*ntimes));
    if( meanwc > 20.0 ){
      lgtotal--;
    }
    singlFflush();
  }
  singlPrintf("Bye!\n");
  MPMY_Finalize();
  exit(0);
}

