/* Flood processor 0 with messages and see how it holds up. */
#include "mpmy.h"
#include "singlio.h"
#include "mpmy_time.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include "protos.h"
#include "Msgs.h"

/* A quick hack to allow calling Sdiag */
FILE *fp;
void printfp(const char *fmt, ...){
  va_list alist;

  va_start(alist, fmt);
  if( fp ){
    vfprintf(fp, fmt, alist);
    fflush(fp);
  }
  va_end(alist);
}

int main(int argc, char **argv){
  int lgmaxlen = 18;
  int lgminlen = 18;
  int lgtotal;
  int lglen;
  int procnum, nproc;
  void *wctmr;
  double wctime;
  void *inbuf[32];
  int i;
  char msgname[512];

  MPMY_Init(&argc, &argv);
  if( argc >= 2) 
    lgminlen = atoi(argv[1]);
  if( argc >= 3 )
    lgmaxlen = atoi(argv[2]);

  wctmr = MPMY_CreateTimer(MPMY_WC_TIME);
  procnum = MPMY_Procnum();
  nproc = MPMY_Nproc();

  for(i=0; i<nproc; i++){
    inbuf[i] = malloc( 1<<lgmaxlen);
  }

  singlPrintf("#Flooding node 0 with %d bytes from %d others\n", 
	      1<<lgmaxlen, nproc-1);
  singlPrintf("# Totlen Pktlength wctime bwidth msecperpack\n");
  singlFflush();
  lgtotal = lgmaxlen;
  for( lglen = lgmaxlen; lglen >= lgminlen; lglen--){
    int len;
    int ntimes;

    MPMY_Sync();
    len = 1<<lglen;
    ntimes = 1<<(lgtotal - lglen);
    if( procnum == 0 ){
      MPMY_ClearTimer(wctmr);
      MPMY_StartTimer(wctmr);
    }
    while( ntimes-- ){
      if( procnum==0 ){
	int pending[32];
	MPMY_Comm_request recv[32];
	int nleft;

	for(i=1; i<nproc; i++){
	  MPMY_Irecv(inbuf[i], len, i, 1001, &recv[i]);
	  pending[i-1] = i;
	}
	nleft = nproc-1;
	/* This loop is sort of an MPMY_WaitAll */
	while(nleft){
	  for(i=0; i<nleft; i++){
	    MPMY_Status stat;
	    int done;
	    done = 0;
	    MPMY_Test(recv[pending[i]], &done, &stat);
	    if( done ){
	      MPMY_Comm_request sendreq;
	      pending[i] = pending[--nleft];
	      /* Send confirmation */
	      MPMY_Isend(&nleft, 4, MPMY_Source(&stat), 1002, &sendreq);
	      MPMY_Wait(sendreq, NULL);
	    }
	  }
	}
      }else{
	MPMY_Comm_request recvreq;
	MPMY_Comm_request sendreq;
	int junk;
	MPMY_Isend(inbuf[0], len, 0, 1001, &sendreq);
	MPMY_Irecv(&junk, 4, 0, 1002, &recvreq);
	MPMY_Wait(sendreq, NULL);
	MPMY_Wait(recvreq, NULL);
      }
    }
    if( procnum == 0 ){
      MPMY_StopTimer(wctmr);
      wctime = MPMY_ReadTimer(wctmr);

      ntimes = 1<<(lgtotal - lglen);
      /* bandwidth is estimated as totalbytes/time.  Latency only counts
	 incoming packets */
      singlPrintf("%9d %9d %#7.3g %#7.3g %#7.3g\n", 
		  (nproc-1)<<lgtotal, len, wctime, 
		  ((nproc-1)<<lgtotal)/((1<<20)*wctime),
		  wctime*1000./(ntimes*(nproc-1)));

      singlFflush();
    }
    /* This acts as a barrier, and also guarantees that everybody is still
       on the same page vis a vis lgtotal */
    MPMY_Bcast(&wctime, 1, MPMY_DOUBLE, 0);
    if( wctime > 20.0 ){
      lgtotal--;
    }
  }
  singlPrintf("#Bye!\n");
  sprintf(msgname, "flood.stats.%d", procnum);
  fp = fopen(msgname, "w");
  MPMY_Diagnostic(printfp);
  exit(0);
}

