/* Like bench.c, but we are sending in both directions simultaneously. */
#include "mpmy.h"
#include "singlio.h"
#include "mpmy_time.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include "protos.h"
#include "error.h"

#define MEGA (1000000.)

void *sendbuf, *recvbuf;

int main(int argc, char **argv){
  int lgmaxlen = 24;
  int lgtotal;
  int lgminlen = 6;
  int lglen;
  int procnum, nproc, partner;
  void *wctmr;
  double wctime;
  double meanwc;
  double maxwc;

  MPMY_Init(&argc, &argv);

  if( argc >= 2) 
    lgminlen = atoi(argv[1]);
  if( argc >= 3 )
    lgmaxlen = atoi(argv[2]);
  wctmr = MPMY_CreateTimer(MPMY_WC_TIME);
  procnum = MPMY_Procnum();
  nproc = MPMY_Nproc();

  if( nproc%2 ){
    printf("Must have an even number of processors\n");
    exit(0);
  }

  sendbuf = malloc( 1<<lgmaxlen );
  recvbuf = malloc( 1<<lgmaxlen );

  if( procnum >= nproc/2 ){
    partner = procnum - nproc/2;
  }else{
    partner = procnum + nproc/2;
  }

  singlPrintf("#Shifting %d bytes in each test\n", 1<<lgmaxlen);
  singlPrintf("# Totlen Pktlength wctime wcmax bwidth(mean) bwidth(min) msec/pack(mean) msec/pack(max)\n");
  singlFflush();
  lgtotal = lgmaxlen;
  for( lglen = lgmaxlen; lglen >= lgminlen; lglen--){
    int len;
    int ntimes;

    MPMY_Sync();
    len = 1<<lglen;
    ntimes = 1<<(lgtotal - lglen);
    MPMY_ClearTimer(wctmr);
    MPMY_StartTimer(wctmr);
    errno=0;
    while( ntimes-- ){
      int ret;
      ret = MPMY_Shift(partner, recvbuf, len, sendbuf, len, NULL);
      if( ret != MPMY_SUCCESS )
	Error("mpmy_shift failed with errno=%d, ret=%d\n", errno, ret);
    }
    MPMY_StopTimer(wctmr);
    wctime = MPMY_ReadTimer(wctmr);
    MPMY_Combine(&wctime, &meanwc, 1, MPMY_DOUBLE, MPMY_SUM);
    MPMY_Combine(&wctime, &maxwc, 1, MPMY_DOUBLE, MPMY_MAX);

    meanwc /= nproc;
    ntimes = 1<<(lgtotal - lglen);
    /* bandwidth is estimated as 2*totalbytes/time, counting both the 
       incoming and outgoing bytes in the Shift.  This is "justified"
       by direct comparison with the half-duplex test.  We only count
       one unit of "latency" here, though */
    singlPrintf("%9d %9d %#7.3g %#7.3g %#7.3g %#7.3g %7.3g %7.3g\n", 
		1<<lgtotal, len, meanwc, maxwc, 
		(2<<lgtotal)/(MEGA*meanwc), 
		(2<<lgtotal)/(MEGA*maxwc), 
		meanwc*1000./ntimes,
		maxwc*1000./ntimes);
    /* Keep things running reasonably quickly */
    if( meanwc  > 20.0 ){
      lgtotal--;
    }
    singlFflush();
  }
  singlPrintf("Bye!\n");
  MPMY_Finalize();
  exit(0);
}

