/* Originally msw's shrink_sdf.c. */

#define NDIM 3
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <values.h>
#include <SDF.h>
#include <assert.h>
#include "macr.h"
#include "vop.h"
#include "fastflpt.h"
#include "randoms.h"

static char outbuf[65536];
static ran_state ranstate;
int seed = 14353;


void
main(int argc, char **argv)
{
    FILE *fpout;
    char outfile[512];
    char hdrfile[512];
    char infile[512];
    int k;
    SDF *inpt;
    int nobj;
    int nout = 0;
    int ret;
    int sfac;
    int p;
    int procnum = 0;
    struct {
	float mass;
	float pos[NDIM];
	float vel[NDIM];
    } b;
    int has_vels = 0;
	

    if (argc == 4) {
	strcpy(infile, argv[1]);
	strcpy(outfile, argv[2]);
	sfac = atoi(argv[3]);
    }
    else {
	fprintf(stderr, 
	   "usage: %s infile outfile shrink\n",
		argv[0]);
	exit(1);
    }
    inpt = SDFopen(NULL, infile);
    if(inpt == NULL){
	fprintf(stderr, "Sorry, couldn't SDFopen %s\n%s\n",
		infile, SDFerrstring);
	inpt = SDFopen("tree.hdr", infile);
	if(inpt == NULL){
	    fprintf(stderr, "Sorry, couldn't SDFopen %s\n%s\n",
		    infile, SDFerrstring);
	    exit(1);
	}
    }
    SDFsetmaxbufsz(65536);
    if( SDFgetint(inpt, "npart", &nobj) ) {
	nobj = SDFnrecs("x", inpt);
    }

    if( SDFhasname("vx", inpt) )
      has_vels = 1;

    Fopen(fpout, outfile, "w");
    setvbuf(fpout, outbuf, _IOFBF, sizeof(outbuf));

    ran_init(seed*(procnum+1), &ranstate);

    p = 2 * sfac;
    
    /* main loop over all particles */
    for(k=0; k<nobj; ){
	if (has_vels) {
	    ret = SDFseekrdvecs(inpt,
				"mass", k, 1, &(b.mass), sizeof(b),
				"x", k, 1, &(b.pos[0]), sizeof(b),
				"y", k, 1, &(b.pos[1]), sizeof(b),
				"z", k, 1, &(b.pos[2]), sizeof(b),
				"vx", k, 1, &(b.vel[0]), sizeof(b),
				"vy", k, 1, &(b.vel[1]), sizeof(b),
				"vz", k, 1, &(b.vel[2]), sizeof(b),
				NULL);
	} else {
	    ret = SDFseekrdvecs(inpt,
				"mass", k, 1, &(b.mass), sizeof(b),
				"x", k, 1, &(b.pos[0]), sizeof(b),
				"y", k, 1, &(b.pos[1]), sizeof(b),
				"z", k, 1, &(b.pos[2]), sizeof(b),
				NULL);
	}
	if (ret) Error("SDFseekrdvecs failed\n");
	b.mass *= sfac;
	if (has_vels) {
	    Fwrite(&b, sizeof(b), 1, fpout);
	} else {
	    Fwrite(&b, sizeof(b)-NDIM*sizeof(float), 1, fpout);
	}
	nout++;
	k += 1+(int)(p * uniform_rand(&ranstate));
    }
    SDFclose(inpt);
    Fclose(fpout);
    sprintf(hdrfile, "%s.hdr", outfile);
    Fopen(fpout, hdrfile, "w");

    fprintf (fpout, "# SDF\n");
    fprintf(fpout, "parameter byteorder = 0x%x;\n", SDFcpubyteorder());
    fprintf(fpout, "int npart = %d;\n", nout);
    fprintf(fpout, "struct {\n");
    fprintf(fpout, "\tfloat mass;\n");
    fprintf(fpout, "\tfloat x;\n");
    fprintf(fpout, "\tfloat y;\n");
    fprintf(fpout, "\tfloat z;\n");
    if (has_vels) {
	fprintf(fpout, "\tfloat vx;\n");
	fprintf(fpout, "\tfloat vy;\n");
	fprintf(fpout, "\tfloat vz;\n");
    }
    fprintf(fpout, "}[%d];\n", nout);
    fprintf(fpout, "#\f\n");
    fprintf(fpout, "# SDF-EOH\n");
    
    Fclose(fpout);

    exit(0);
}
