/* Like msw's shrink_sdf.c, but use ReadBodies. */

/* 
   This program can now read from a pipe.   Here's a line from an ftp
   session to unitree.  For obscure reasons, the hdr file must be separate
   from the data in order to read from a pipe.  This is broken because
   the data file has a header too
ftp>  get cdm_c.0350 "|/home/coho/johns/t16.3/SDFshrink/SDFshrinkrb.sun4 --hdr-in hdr.0350 --out cdm_c.0650.20th -fac 20

*/
#define NDIM 3
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <values.h>
#include <stddef.h>
#include "getopt.h"
#include "Malloc.h"
#include "SDF.h"
#include "Assert.h"
#include "macr.h"
#include "vop.h"
#include "fastflpt.h"
#include "randoms.h"
#include "readbodies.h"
#include "protos.h"

static char outbuf[65536];
static ran_state ranstate;
int seed = 14353;

#define Map(i, Name, member) do{\
  na_arr[i].name = Name;\
  na_arr[i].offset = offsetof(body, member);\
}while(0)
struct name_address_s na_arr[7];

FILE *fpout;
float sfac_inverse;
float sfac;
int has_vels = 0;
int nout = 0;

typedef struct {
    float mass;
    float pos[NDIM];
    float vel[NDIM];
} body;

void DoBody(body *b);

enum {
    OUTPUT_OPT=9999, INPUT_OPT, HDRINPUT_OPT, HDROUTPUT_OPT, FACTOR_OPT};

struct option optab[] = {
    {"output", required_argument, NULL, OUTPUT_OPT},
    {"input", required_argument, NULL, INPUT_OPT},
    {"hdr-input", required_argument, NULL, HDRINPUT_OPT},
    {"hdr-output", required_argument, NULL, HDROUTPUT_OPT},
    {"factor", required_argument, NULL, FACTOR_OPT},
    {"no-vel", no_argument, &has_vels, 0},
    {"vel", no_argument, &has_vels, 1},
    {NULL, 0, NULL, 0}
};

void
main(int argc, char **argv)
{
    char *outfile = NULL;
    char *hdrout = NULL;
    char *infile = NULL;
    int nnames;
    int c;

    has_vels = 1;
    for(optopt=0;
	(c=getopt_long_only(argc, argv, "", optab, NULL)) != EOF;
	optopt=0){
	switch(c){
	case 0:
	    /* a long-option with a non-NULL flag */
	    break;
	case INPUT_OPT:
	    infile = optarg;
	    break;
	case OUTPUT_OPT:
	    outfile = optarg;
	    break;
	case HDRINPUT_OPT:
	    ReadBodiesSDFhdr = optarg;
	    break;
	case HDROUTPUT_OPT:
	    ReadBodiesSDFhdr = optarg;
	    break;
	case FACTOR_OPT:
	    sfac = atof(optarg);
	    break;
	}
    }
    
    if( infile == NULL ){
	infile = "-";
    }
    if( outfile == NULL ){
	outfile = "-";
    }
    if( hdrout == NULL ){
	hdrout = Malloc(strlen(outfile) + 20);
	sprintf(hdrout, "%s.hdr", outfile);
    }

    if( sfac < 1.0 ){
	fprintf(stderr, "Sorry, sfac must be >=1.0\n");
	exit(1);
    }
    sfac_inverse = 1.0F/sfac;

    if( has_vels ){
	Map(0, "mass", mass);
	Map(1, "x", pos[0]);
	Map(2, "y", pos[1]);
	Map(3, "z", pos[2]);
	Map(4, "vx", vel[0]);
	Map(5, "vy", vel[1]);
	Map(6, "vz", vel[2]);
	nnames = 7;
    }else{
	Map(0, "mass", mass);
	Map(1, "x", pos[0]);
	Map(2, "y", pos[1]);
	Map(3, "z", pos[2]);
	nnames=4;
    }

    Fopen(fpout, outfile, "w");
    setvbuf(fpout, outbuf, _IOFBF, sizeof(outbuf));

    ran_init(seed, &ranstate);

    /* main loop over all particles */
    ReadBodies(infile, nnames, sizeof(body), na_arr, DoBody);
    Fclose(fpout);
    Fopen(fpout, hdrout, "w");

    fprintf (fpout, "# SDF\n");
    fprintf(fpout, "parameter byteorder = 0x%x;\n", SDFcpubyteorder());
    fprintf(fpout, "int npart = %d;\n", nout);
    fprintf(fpout, "struct {\n");
    fprintf(fpout, "\tfloat mass;\n");
    fprintf(fpout, "\tfloat x;\n");
    fprintf(fpout, "\tfloat y;\n");
    fprintf(fpout, "\tfloat z;\n");
    if (has_vels) {
	fprintf(fpout, "\tfloat vx;\n");
	fprintf(fpout, "\tfloat vy;\n");
	fprintf(fpout, "\tfloat vz;\n");
    }
    fprintf(fpout, "}[%d];\n", nout);
    fprintf(fpout, "#\f\n");
    fprintf(fpout, "# SDF-EOH\n");
    
    Fclose(fpout);

    exit(0);
}

void DoBody(body *b)
{
    if( uniform_rand(&ranstate) > sfac_inverse )
	return;
    b->mass *= sfac;
    if (has_vels) {
	Fwrite(b, sizeof(body), 1, fpout);
    } else {
	Fwrite(b, sizeof(body)-NDIM*sizeof(float), 1, fpout);
    }
    nout++;
    if( nout %1000 == 0 )
	putc('k', stderr);
}
